/*----------------------------------------------------------------------------*/
/* Xymon message combiner tool.                                               */
/*                                                                            */
/* This is used to send multiple files (or directories thereof), each         */
/* containing a separate Xymon message, to a Xymon/BB server using the        */
/* extcombine format.                                                         */
/*                                                                            */
/* Copyright (C) 2002-2015 Henrik Storner <henrik@hswn.dk>                    */
/*                                                                            */
/* This program is released under the GNU General Public License (GPL),       */
/* version 2. See the file "COPYING" for details.                             */
/*                                                                            */
/*----------------------------------------------------------------------------*/

static char rcsid[] = "$Id: xymoncombine.c 7223 2013-07-25 19:56:20Z storner $";

#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include <ctype.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <errno.h>
#include <limits.h>

#include "libxymon.h"


/* Some safety mechanisms here */
#define MAX_SIZE 4*1024*1024
#define MAX_DEPTH 4



/* Send it in to combo_add */
void handle_file(char *filename) {
	FILE *fd;
	size_t datasize;
        struct stat st;
	static char *msgbuf;
	size_t n;

	/* static */
	if (msgbuf == NULL) msgbuf = (char *)malloc(MAX_SIZE);

	dbgprintf(" -> handle_file %s\n", filename);

	if (! (stat(filename, &st) == 0) ) {
		errprintf("Error stating file '%s': %s\n", filename, strerror(errno));
		return;
	}

	if (! S_ISREG(st.st_mode)) { dbgprintf("File '%s' not a regular file?\n", filename); return; }

	if (st.st_size > (MAX_SIZE - 10240)) {
		errprintf("File '%s' too large (%zu bytes); skipping\n", filename, st.st_size);
		return;
	}		

	fd = fopen(filename, "r");
	if (!fd) {
		errprintf("Error opening file '%s': %s\n", filename, strerror(errno));
		return;
	}

	n = fread(msgbuf, 1, st.st_size, fd);
	fclose(fd);

	if (n != st.st_size) errprintf("'%s' not all data read (size=%zu, data=%zu)\n", filename, st.st_size, n);

	*(msgbuf+n) = '\0';
	combo_addcharbytes(msgbuf, n);

	*msgbuf = '\0';

	dbgprintf(" <- handle_file %s\n", filename);
}



/* Most of this from lib/files.c */
void handle_directory(char *dirname, int nesting) {

	DIR *dirfd;
        struct dirent *de;
        struct stat st;
        char fn[PATH_MAX];

	dbgprintf(" -> handle_directory/%d %s\n", nesting, dirname);

	if (nesting++ > MAX_DEPTH) {
		errprintf("Maximum recursion level reached! Backing out!\n");
		return;
	}

	dirfd = opendir(dirname);
	if (!dirfd) {
		errprintf("Error opening directory %s: %s\n", dirname, strerror(errno));
		return;
	}

	/* Recurse down */
	while ( (de = readdir(dirfd)) != NULL ) {
	        sprintf(fn, "%s/%s", dirname, de->d_name);
	        if (strcmp(de->d_name, ".") && strcmp(de->d_name, "..") && (stat(fn, &st) == 0)) {
			if (S_ISREG(st.st_mode)) handle_file(fn);
			else if (S_ISDIR(st.st_mode)) handle_directory(fn, nesting);
			else errprintf("'%s' not directory or regular file; ignoring\n", fn);
		}
	}
	closedir(dirfd);

	dbgprintf(" <- handle_directory/%d\n", nesting);
	return;
}




int main(int argc, char *argv[])
{
	int timeout = (getenv("XYMON_TIMEOUT") != NULL) ? atoi(xgetenv("XYMON_TIMEOUT")) : XYMON_TIMEOUT;
	int result = 1;
	int argi;
	int showhelp = 0;
	char *recipient = NULL;
	strbuffer_t *msg = newstrbuffer(0);
	FILE *respfd = stdout;
	char *envarea = NULL;
	sendreturn_t *sres;
	int filesystemargs = 0;
	struct stat args_st;
	int wantresponse = 0, mergeinput = 0, usebackfeedqueue = 0;

	for (argi=1; (argi < argc); argi++) {
		if (strcmp(argv[argi], "--debug") == 0) {
			debug = 1;
		}
		else if (strncmp(argv[argi], "--proxy=", 8) == 0) {
			char *p = strchr(argv[argi], '=');
			setproxy(p+1);
		}
		else if (strcmp(argv[argi], "--help") == 0) {
			showhelp = 1;
		}
		else if (strcmp(argv[argi], "--version") == 0) {
			fprintf(stdout, "Xymon version %s\n", VERSION);
			return 0;
		}
		else if (strcmp(argv[argi], "--str") == 0) {
			respfd = NULL;
		}
		else if (strncmp(argv[argi], "--out=", 6) == 0) {
			char *fn = argv[argi]+6;
			respfd = fopen(fn, "wb");
		}
		else if (strncmp(argv[argi], "--timeout=", 10) == 0) {
			char *p = strchr(argv[argi], '=');
			timeout = atoi(p+1);
		}
		else if (argnmatch(argv[argi], "--env=")) {
			char *p = strchr(argv[argi], '=');
			loadenv(p+1, envarea);
		}
		else if (argnmatch(argv[argi], "--area=")) {
			char *p = strchr(argv[argi], '=');
			envarea = strdup(p+1);
		}
		else if (argnmatch(argv[argi], "--compress")) {
			char *p = strchr(argv[argi], '=');
			enablecompression = 1;
			if (!p || (++p == NULL)) { dbgprintf("Compressing using default library\n"); }
			else {
				defaultcompression = parse_compressiontype(p);
				if (defaultcompression == COMP_UNKNOWN && getenv("XYMON_COMPTYPE")) defaultcompression = parse_compressiontype(getenv("XYMON_COMPTYPE"));
				if (defaultcompression == COMP_UNKNOWN) {
					errprintf("Requested compression type '%s' not recognized\n", p);
					return 1;
				}
				dbgprintf("Compressing using '%s'\n", comptype2str(defaultcompression));
			}
		}
		else if (argnmatch(argv[argi], "--no-compress")) {
			dbgprintf("Compression disabled\n");
			enablecompression = -1;
		}
		else if (strncmp(argv[argi], "--bfq", 5) == 0) {
			usebackfeedqueue = 1;
			if (strncmp(argv[argi], "--bfq=", 6) == 0) {
				char *p = strchr(argv[argi], '=');
				backfeedqueuenumber = atoi(p+1);
				if ((backfeedqueuenumber < 0) || (backfeedqueuenumber > 9)) {
					errprintf("Invalid BFQ channel: %d\n", backfeedqueuenumber);
					return -1;
				}
				else dbgprintf("Going to use BFQ number: %d\n", backfeedqueuenumber);
			}
			/* else, backfeedqueuenumber = default */
		}
		else if (strcmp(argv[argi], "-?") == 0) {
			showhelp = 1;
		}
		else if ((*(argv[argi]) == '-') && (strlen(argv[argi]) > 1)) {
			fprintf(stderr, "Unknown option %s\n", argv[argi]);
		}
		else {
			/* No more options - pickup recipient and/or file/directory names */
			if (recipient == NULL) {
				recipient = argv[argi];
			}
			else if (!filesystemargs) {
				/* Record which argument begins the file or directory arguments */
				filesystemargs = argi;
				dbgprintf("Assuming files/dirs starting at arg %d\n", filesystemargs);
			}
		}
	}

	if (recipient == NULL) {
		if (getenv("XYMSRV") && (strcmp(xgetenv("XYMSRV"), "") != 0)) recipient = xgetenv("XYMSRV");
		else if (getenv("XYMSERVERS") && (strcmp(xgetenv("XYMSERVERS"), "") != 0)) recipient = "0.0.0.0";
		else if (usebackfeedqueue) {
			// recipient =  (char *)malloc(22);			/* future use */
			// sprintf(recipient, "%d", backfeedqueuenumber);	/* future use */
			recipient = "0";
			fprintf(stderr, "No recipient specified but bfq given (%d)\n", backfeedqueuenumber);
		}
		else {
			recipient = "127.0.0.1";
			fprintf(stderr, "No recipient specified - assuming localhost:1984\n");
		}
	}

	if ((recipient == NULL) || (filesystemargs < 1) || showhelp) {
		fprintf(stderr, "Usage: %s [options] RECIPIENT <file | directory> [ <file | directory> ]\n", argv[0]);
		fprintf(stderr, "  RECIPIENT: IP address, hostname, or URL,\n");
		fprintf(stderr, "    or '0' for BFQ, or '0.0.0.0' to read from $XYMONSERVERS\n");
		return 1;
	}


	/* We're combining, so you're not getting a response */
	wantresponse = 0;


	/* If we're using a backfeed queue, set it up */
	usebackfeedqueue = (strcmp(recipient, "0") == 0) ? (sendmessage_init_local() > 0) : 0;
	if (usebackfeedqueue) combo_start_local(); else combo_start();


	/* Loop through all remaining arguments, which could be either files or directories */
	for (argi = filesystemargs; (argi < argc); argi++) {

		if (stat(argv[argi], &args_st) == 0) {
                        if (S_ISREG(args_st.st_mode)) handle_file(argv[argi]);
                        else if (S_ISDIR(args_st.st_mode)) handle_directory(argv[argi], 1);
                        else errprintf("'%s' not directory or regular file; ignoring\n", argv[argi]);
                }
		else errprintf("Error reading '%s': %s\n", argv[argi], strerror(errno));
	}


	/* Clean up */
	combo_end();
        if (usebackfeedqueue) sendmessage_finish_local();

	return 0;
}

